import java.util.Vector;

public class PhysicalRobotServer implements RobotServer {
  public static final char SERVER_ID = 0;

  public static long TIMEOUT = 1000l;
  public static long CHARACTER_DELAY = 0l;
  public static long PACKET_DELAY = 10l;
  public static long ERROR_DELAY = 25l;
  public static long POLL_INTERVAL = 5000l;
  public static int DEBUG = 0;

  private ServerGUI gui;
  private Vector robots;
  private Vector robot_packet_queue;
  private class PacketQueueElement {
    public Packet packet;
    public char source;
  }

  private RobotLink robot_link;
  private LocalClientLink local_client_link;
  private RobotClient local_client;

  public void setGUI(ServerGUI gui) {
    if(DEBUG>0)
      System.out.println("Setting GUI ...");
    this.gui = gui;
  }

  public Vector getRobots(boolean all_robots) {
    if(!all_robots) {
      //return only those robots that are available (ready and not owned)
      Vector available_robots = new Vector();
      for(int i=0; i<robots.size(); i++)
        if(((Robot)robots.elementAt(i)).isReady() && !((Robot)robots.elementAt(i)).isOwned())
          available_robots.add(robots.elementAt(i));
      return available_robots;
    } else
      return robots;
  }

  public Robot getRobot(char robot_id) {
    Robot robot;
    for(int i=0; i<robots.size(); i++) {
      robot = (Robot)robots.elementAt(i);
      if(robot.getID()==robot_id)
        return robot;
    }
    return null;
  }

  public LocalClientLink getLocalClientLink() {
    return local_client_link;
  }

  public void startLocalClient(Robot robot) {
    if(local_client==null || !local_client.running) {
      local_client = new HumanRobotClient(robot);
      local_client.setUserInterface(new ClientGUI());
      local_client.setServerLink(new LocalServerLink(local_client_link));
      local_client.start();
    }
  }

  public void pollForRobots() {
    if(DEBUG>0)
      System.out.println("Polling for robots ...");
    Packet packet = new Packet(Opcode.HEADER, SERVER_ID, Robot.ANY_ROBOT_ID, (char)0, Opcode.REQUEST_NEW_ROBOT_IDS, (char)0, null, (char)0);
    Packet response = robot_link.getResponse(packet);
    if(response!=null && response.opcode==Opcode.ACKNOWLEDGEMENT) {
      char robot_id = response.source_id;

      if(DEBUG>0)
        System.out.println("Found robot #"+(int)robot_id);

      //acknowledge robot
      packet = new Packet(Opcode.HEADER, SERVER_ID, robot_id, (char)0, Opcode.ACKNOWLEDGEMENT, (char)0, null, (char)0);
      robot_link.sendPacket(packet);

      PhysicalRobot robot = new PhysicalRobot(robot_id);
      addRobot(robot);
      try{Thread.sleep(PACKET_DELAY);}catch(Throwable t){}
    } else
      try{Thread.sleep(ERROR_DELAY);}catch(Throwable t){}
  }

  private void addRobot(Robot robot) {
    if(getRobot(robot.getID())==null) {
      robots.add(robot);
      robot_packet_queue.add(new Vector());
      if(gui!=null)
        gui.addRobot(robot);
    } else
      System.out.println("Warning: Robot #"+(int)robot.getID()+" already added to list.");
  }

  public boolean connectRobot(Robot robot) {
    if(robot!=null && !robot.isOwned()) {
      robot.setOwned(true);
      return true;
    } else
      return false;
  }

  public void disconnectRobot(Robot robot) {
    if(robot!=null)
      robot.setOwned(false);
  }

  private void removeRobot(Robot robot) {
    robots.remove(robot);
    gui.removeRobot(robot);
  }

  private void queuePacket(Packet packet, char source) {
    for(int i=0; i<robots.size(); i++) {
      if(((Robot)robots.elementAt(i)).getID()==packet.destination_id){
        PacketQueueElement pqe = new PacketQueueElement();
        pqe.packet = packet;
        pqe.source = source;
        ((Vector)robot_packet_queue.elementAt(i)).add(pqe);
        return;
      }
    }    
  }

  public void setSense(char client_id, char robot_id, boolean sense) {
    char[] _sense = new char[1];
    if(sense)
      _sense[0] = 1;
    else
    	_sense[0] = 0;
    Packet packet = new Packet(Opcode.HEADER, SERVER_ID, robot_id, (char)0, Opcode.SET_SENSE, (char)1, _sense, (char)0);
    queuePacket(packet, client_id);  
  }

  public void setAutoScan(char client_id, char robot_id, boolean auto_scan) {
    char[] autoscan = new char[1];
    if(auto_scan)
      autoscan[0] = 1;
    else
    	autoscan[0] = 0;
    Packet packet = new Packet(Opcode.HEADER, SERVER_ID, robot_id, (char)0, Opcode.SET_AUTO_SCAN, (char)1, autoscan, (char)0);
    queuePacket(packet, client_id);
  }

  public void scanOnce(char client_id, char robot_id) {
    Packet packet = new Packet(Opcode.HEADER, SERVER_ID, robot_id, (char)0, Opcode.SCAN, (char)0, null, (char)0);
    queuePacket(packet, client_id);
  }

  public void moveRobot(char client_id, char robot_id, char direction, char distance) {
    char[] move_command = new char[2];
    move_command[0] = direction;
    move_command[1] = distance;
    Packet packet = new Packet(Opcode.HEADER, SERVER_ID, robot_id, (char)0, Opcode.MOVE, (char)2, move_command, (char)0);
    queuePacket(packet, client_id);
  }

  public void stopRobot(char client_id, char robot_id) {
    Packet packet = new Packet(Opcode.HEADER, SERVER_ID, robot_id, (char)0, Opcode.STOP, (char)0, null, (char)0);
    queuePacket(packet, client_id);
  }

  public void updateRobotInfo(Robot robot) {
    if(DEBUG>0)
      System.out.println("Robot #"+(int)robot.getID()+": Updating Info ...");

    Packet packet = new Packet(Opcode.HEADER, SERVER_ID, robot.getID(), (char)0, Opcode.REQUEST_INFO, (char)0, null, (char)0);
    Packet response = robot_link.getResponse(packet);
    if(response!=null && response.opcode==Opcode.DATA && response.data!=null) {
      char[] info = new char[response.length];
      for(int i=0; i<response.length; i++)
        info[i]=response.data[i];
      if(DEBUG>0) {
        System.out.print("      ");
        System.out.print((int)response.data[0]);
        for(int j=1; j<response.length; j++)
          System.out.print(response.data[j]);
        System.out.println();
      }
      robot.setInfo(new RobotInfo(info));
      try{Thread.sleep(PACKET_DELAY);}catch(Throwable t){}
    } else
      try{Thread.sleep(ERROR_DELAY);}catch(Throwable t){}
  }

  public void updateRobotSensorLayout(Robot robot) {
    if(DEBUG>0)
      System.out.println("Robot #"+(int)robot.getID()+": Updating Sensor Layout ...");
    Packet packet = new Packet(Opcode.HEADER, SERVER_ID, robot.getID(), (char)0, Opcode.REQUEST_SENSOR_LAYOUT, (char)0, null, (char)0);
    Packet response = robot_link.getResponse(packet);
    if(response!=null && response.opcode==Opcode.DATA && response.length%2==0 && response.data!=null) {
      char[] layout = new char[response.length];
      for(int i=0; i<response.length; i++)
        layout[i]=response.data[i];
      if(DEBUG>0) {
        System.out.print("      ");
        for(int j=0; j<response.length; j++)
          System.out.print((int)response.data[j]+" ");
        System.out.println();
      }
      robot.setSensorLayout(layout);
      try{Thread.sleep(PACKET_DELAY);}catch(Throwable t){}
    } else
      try{Thread.sleep(ERROR_DELAY);}catch(Throwable t){}
  }

  public void updateRobotSensorData(Robot robot) {
    if(DEBUG>0)
      System.out.println("Robot #"+(int)robot.getID()+": Updating Sensor Data ...");

    //get regular sensor data
    Packet packet = new Packet(Opcode.HEADER, SERVER_ID, robot.getID(), (char)0, Opcode.REQUEST_SENSOR_DATA, (char)0, null, (char)0);
    Packet response = robot_link.getResponse(packet);
    if(response!=null && response.opcode==Opcode.DATA && response.data!=null) {
      if(DEBUG>0){
        System.out.println(response.toString());
        System.out.println();
      }
      char[] data = new char[response.length];
      for(int i=0; i<response.length; i++)
        data[i]=response.data[i];
      robot.setSensorData(data);
      try{Thread.sleep(PACKET_DELAY);}catch(Throwable t){}
    } else
      try{Thread.sleep(ERROR_DELAY);}catch(Throwable t){}

    if(robot.hasSonarSensor()) {
      if(DEBUG>0)
        System.out.println("Robot #"+(int)robot.getID()+": Updating Sonar Data ...");

      //get sonar data
      packet = new Packet(Opcode.HEADER, SERVER_ID, robot.getID(), (char)0, Opcode.REQUEST_SONAR_DATA, (char)0, null, (char)0);
      response = null;
      response = robot_link.getResponse(packet);
      if(response!=null && response.opcode==Opcode.DATA && response.data!=null) {
        if(DEBUG>0) {
          System.out.println(response.toString());
          System.out.println();
        }
        robot.setSonarData(response.data);

        try{Thread.sleep(PACKET_DELAY);}catch(Throwable t){}
      } else
        try{Thread.sleep(ERROR_DELAY);}catch(Throwable t){}
    }
  }

  private void updateRobots() {
    Robot robot;
    for(int i=0; i<robots.size(); i++) {
      robot = (Robot)robots.elementAt(i);
      if(robot.getInfo()==null) {
        updateRobotInfo(robot);
        continue;
      }
      if(robot.getSensors()==null) {
        updateRobotSensorLayout(robot);
        continue;
      }
      if(robot.isSensing())
        updateRobotSensorData(robot);

      //send queued packet
      int robot_index = robots.indexOf(robot);
      if(robot_index>=0) {
        Vector pqes = (Vector)robot_packet_queue.elementAt(robot_index);
        if(pqes!=null && pqes.size()>0) {
          PacketQueueElement pqe = (PacketQueueElement)pqes.elementAt(0);
          if(DEBUG>0)
            System.out.println("Sending "+Opcode.stringValue(pqe.packet.opcode)+" to Robot #"+(int)robot.getID());
          Packet response = robot_link.getResponse(pqe.packet);
          if(response!=null && response.opcode==Opcode.ACKNOWLEDGEMENT) {
            if(pqe.packet.opcode==Opcode.SET_SENSE)
              getRobot(pqe.packet.destination_id).setSensing(pqe.packet.data[0]!=0);
            pqes.removeElementAt(0);
            //send ack to client
          } else {
            //leave packet in queue to try again
            //send neg ack to client
          }
        } else {
          if(DEBUG>1)
            System.out.print("Robot #"+(int)robot.getID()+": Ping ...");
          Packet packet = new Packet(Opcode.HEADER, SERVER_ID, robot.getID(), (char)0, Opcode.NO_OP, (char)0, null, (char)0);
          Packet response = robot_link.getResponse(packet);
          if(DEBUG>1){
            if(response!=null && response.opcode==Opcode.NO_OP)
              System.out.println(" Pong");
            else
              System.out.println();
          }
        }
      }
    }
  }

  private void init() {
    if(gui!=null)
      gui.setVisible(true);

    robots = new Vector();
    robot_packet_queue = new Vector();
    robot_link = new RobotLink(CHARACTER_DELAY, TIMEOUT, DEBUG>1);
    local_client_link = new LocalClientLink(this);
    local_client = null;
  }

  public void run() {
    init();

    long interval_start = System.currentTimeMillis() - POLL_INTERVAL;

    while(true) {
      if(System.currentTimeMillis() - interval_start > POLL_INTERVAL) {
        pollForRobots();
        interval_start = System.currentTimeMillis();
      }
      updateRobots();
      if(gui!=null)
        gui.update();
    }
  }
}

