import java.io.*;
import java.util.*;
import javax.comm.*;

class Serial {
  public static long TIMEOUT = 500;
  public static boolean DEBUG;
  public static long DELAY;

  private InputStream in;
  private OutputStream out;
 
  private SerialPort serialPort = null;

  public Serial(){
    this(null);
  }

  public Serial(String port) {
    CommPortIdentifier portId;

    if(port!=null){ //if port is given, try to open that port
      try {
        portId = CommPortIdentifier.getPortIdentifier(port);
        serialPort = (SerialPort)portId.open("hbserver",6969);
      } catch (NoSuchPortException e) {
        error("No such port: "+port);
      } catch (PortInUseException e) {
        error("Port "+port+" in use");
      } catch (Exception e) {
       error("Error opening streams over serial line");
      };
    } else { //otherwise, search for an available port
      //get a list of all available communications ports
      Enumeration portList;
      portList=CommPortIdentifier.getPortIdentifiers();
      while(portList.hasMoreElements()){
        portId=(CommPortIdentifier)portList.nextElement();
        if(portId.getPortType()==CommPortIdentifier.PORT_SERIAL){
          try{
            System.out.println("Trying " + portId.getName() + " ...");
            serialPort = (SerialPort)portId.open("hbserver",6969);
            break;
          } catch (Exception e) {
            //skip this comm port and try next
            System.err.println(e.getMessage());
            serialPort = null;
          }
        }
      }
    }

    if(serialPort==null)
      error("No ports available");
    else {
      try{
        in = serialPort.getInputStream();
        out = serialPort.getOutputStream();
        serialPort.setSerialPortParams(9600,SerialPort.DATABITS_8,SerialPort.STOPBITS_1,SerialPort.PARITY_NONE);
        System.out.println("Connected to " + serialPort.getName());
      } catch (IOException ioe) {
        error(ioe.getMessage());
      } catch (UnsupportedCommOperationException ucoe) {
        error("Error opening communications stream");
      }
    }
  }

    public void setTimeout(long timeout) {
        this.TIMEOUT = timeout;
    }

  public void setDebug(boolean debug) {
    this.DEBUG = debug;
  }

  public void setDelay(long delay) {
    this.DELAY = delay;
  }

  public void send(char c) {
    try {
      out.write(c);
      out.flush();
    } catch (IOException e) {
      error("Error while sending data over serial line");
    }
  }

  public synchronized boolean sendPacket(Packet packet){
    if(DEBUG) System.out.println("Sending ...");
    if(packet==null) return false;

    try{
      if(DEBUG) System.out.println("pre:"+(int)packet.preamble);
      send(packet.preamble);
      Thread.sleep(DELAY);

      if(DEBUG) System.out.println("src:"+(int)packet.source_id);
      send(packet.source_id);
      Thread.sleep(DELAY);

      if(DEBUG) System.out.println("dst:"+(int)packet.destination_id);
      send(packet.destination_id);
      Thread.sleep(DELAY);

      if(DEBUG) System.out.println("pkt:"+(int)packet.packet_id);
      send(packet.packet_id);
      Thread.sleep(DELAY);

      if(DEBUG) System.out.println("opc:"+(int)packet.opcode);
      send(packet.opcode);
      Thread.sleep(DELAY);

      if(DEBUG) System.out.println("len:"+(int)packet.length);
      send(packet.length);
      Thread.sleep(DELAY);

      if(DEBUG) System.out.print("dat:{ ");
      if(packet.length > 0 && packet.data!=null)
      for(int i=0; i<packet.data.length; i++){
        if(DEBUG) System.out.print((int)packet.data[i]+" ");
        send(packet.data[i]);
        Thread.sleep(DELAY);
      }
      if(DEBUG) System.out.println("}");

      if(DEBUG) System.out.println("crc:"+(int)packet.crc);
      send(packet.crc);
    } catch (Exception e) {
      error("Error while sending data over serial line");
    }

    if(DEBUG) System.out.println("... Done");

    return true;
  }

  public char receive() {
    char temp = Opcode.NO_OP;
    try {
      temp=(char)in.read();
    } catch (IOException e) {
      error(e.getMessage());
      //error("Error while receiving data over serial line");
    }
    return temp;
  }

  public boolean receivePacket(Packet packet) {
    Character c;

    // wait for header
    if(DEBUG) System.out.print("pre:");
    c = receiveTimeout();
    while(c!=null && c.charValue()!=Opcode.HEADER)
      c = receiveTimeout();
    if(c==null)
      return false;
    else
      packet.preamble = c.charValue();
    if(DEBUG) System.out.println((int)packet.preamble);

    // get source id
    if(DEBUG) System.out.print("src:");
    c = receiveTimeout();
    if(c==null)
      return false;
    else
      packet.source_id = c.charValue();
    if(DEBUG) System.out.println((int)packet.source_id);

    // get destination id
    if(DEBUG) System.out.print("dst:");
    c = receiveTimeout();
    if(c==null)
      return false;
    else
      packet.destination_id = c.charValue();
    if(DEBUG) System.out.println((int)packet.destination_id);

    // get packet id
    if(DEBUG) System.out.print("pkt:");
    c = receiveTimeout();
    if(c==null)
      return false;
    else
      packet.packet_id = c.charValue();
    if(DEBUG) System.out.println((int)packet.packet_id);

    // get opcode
    if(DEBUG) System.out.print("opc:");
    c = receiveTimeout();
    if(c==null)
      return false;
    else
      packet.opcode = c.charValue();
    if(DEBUG) System.out.println((int)packet.opcode);

    // get length
    if(DEBUG) System.out.print("len:");
    c = receiveTimeout();
    if(c==null)
      return false;
    else
      packet.length = c.charValue();
    if(DEBUG) System.out.println((int)packet.length);

    if(packet.length >= 0)
      packet.data = new char[packet.length];
    else
      return false;

    // get data
    if(DEBUG) System.out.print("dat: { ");
    for(int i=0; i<packet.length; i++){
      c = receiveTimeout();
      if(c==null)
	return false;
      else
	packet.data[i] = c.charValue();
      if(DEBUG) System.out.print((int)packet.data[i]);
      if(DEBUG) System.out.print(" ");
    }
    if(DEBUG) System.out.println("}");

    // get crc
    if(DEBUG) System.out.print("crc:");
    c = receiveTimeout();
    if(c==null)
      return false;
    else
      packet.crc = c.charValue();
    if(DEBUG) System.out.println(packet.crc);

    return true;
  }

  public int available() {
    try {
      return in.available();
    } catch (IOException e) {
      error("Error while checking availability of data over serial line");
    };
    return 0;
  }

  public Character receiveTimeout(){
    long start_time = System.currentTimeMillis();
    while(available()==0){
      if(System.currentTimeMillis()-start_time > TIMEOUT){
        System.out.println("TIMEOUT!!!");
        return null;
      }
    }
    return new Character(receive());
  }

  public void stop() {
    serialPort.close(); 
  }

  public void error(String s) {
    System.err.println(s);
    System.exit(1);
  }
}
