#ifndef bool
#define bool int
#define true 1
#define false 0
#endif

#define DEBUG 0

#define DELAY 0l
#define TIMEOUT 500l
#define DELAY_PER_DEGREE 0.005556

/* MACROS */
#define disable_pcode_serial() poke(0x3c, 1) /* necessary to receive characters using serial_getchar */
#define enable_pcode_serial() poke(0x3c, 0) /* necessary for IC to interact with board again */

#define LEFT_ENCODER_PORT             0
#define RIGHT_ENCODER_PORT            1
#define BREAK_BEAM_SENSOR_PORT        2
#define LIGHT_SENSOR_PORT             3
#define LEFT_FRONT_BUMP_SENSOR_PORT  14
#define RIGHT_FRONT_BUMP_SENSOR_PORT 15


#define THIS_ROBOT_ID 42
#define ANY_ROBOT_ID 0

#define SENSOR_LAYOUT_SIZE 10

/* type, location */
persistent char sensor_layout[SENSOR_LAYOUT_SIZE]=
{BUMP_SENSOR, LEFT_FRONT,
 BUMP_SENSOR, RIGHT_FRONT,
 BREAK_BEAM_SENSOR, FRONT,
 LIGHT_SENSOR, FRONT,
 SONAR_SENSOR, TOP_FRONT};

#define INFO_SIZE 9
#define SONAR_DATA_SIZE  61
#define SONAR_DATA_START_INDEX  4
#define SONAR_DATA_END_INDEX   64
#define NAME "Johnny 5"

#define SENSOR_DATA_SIZE 65
char sensor_data[SENSOR_DATA_SIZE];

/* flags */
bool registered;
bool sense_enabled;

void init(){
  registered = false;
  sense_enabled = false;
  MIN_SERVO_WAVETIME = 600;
  MAX_SERVO_WAVETIME = 4380;
  servo_on();
  servo_deg(0.0);
  msleep(1000l);
  sonar_init();
  disable_pcode_serial();
  initSerial();
}

void close(){
  servo_off();
  closeSerial();
  enable_pcode_serial();
}

void waitFor(char c){
  while(!dataAvailable() || serialGetChar()!=c){
    defer();
  }
}

void enable_read(){
  disable_pcode_serial();
}

void disable_read(){
  int i;
  enable_pcode_serial();
  /* clear out buffer */
  for(i=0; i<BUFFER_SIZE; i++)
    buffer[i] = 0;
}

bool receivePacket(){
  char index;
  long start_time;

  /* get header */
  waitFor(HEADER);
  packet_buffer[PREAMBLE_INDEX] = HEADER;

  /* get source id */
  packet_buffer[SOURCE_ID_INDEX] = serialGetChar();

  /* get destination id */
  packet_buffer[DESTINATION_ID_INDEX] = serialGetChar();
  if(packet_buffer[DESTINATION_ID_INDEX] != THIS_ROBOT_ID &&
     packet_buffer[DESTINATION_ID_INDEX] != ANY_ROBOT_ID)
    return false;
      
  /* get packet id */
  packet_buffer[PACKET_ID_INDEX] = serialGetChar();

  /* get opcode */
  packet_buffer[OPCODE_INDEX] = serialGetChar();

  /* get length */
  packet_buffer[LENGTH_INDEX] = serialGetChar();
  if(packet_buffer[LENGTH_INDEX] < 0 || packet_buffer[LENGTH_INDEX] > MAX_DATA_SIZE)
    return false;

  /* get data */
  for(index=0; index<packet_buffer[LENGTH_INDEX]; index++){
    start_time=mseconds();
    while(!dataAvailable()){
      if(mseconds()-start_time > TIMEOUT)
	return false;
      else
	defer();
    }
    packet_buffer[DATA_INDEX+index]=serialGetChar();
  }

  /* get crc */
  packet_buffer[CRC_INDEX] = serialGetChar();
  return true;
}

bool sendPacket(){
  char index;

  if(packet_buffer[LENGTH_INDEX] < 0 || packet_buffer[LENGTH_INDEX] > MAX_DATA_SIZE)
    return false;

  serialPutChar(packet_buffer[PREAMBLE_INDEX]);
  msleep(DELAY);

  serialPutChar(THIS_ROBOT_ID);
  msleep(DELAY);

  serialPutChar(packet_buffer[DESTINATION_ID_INDEX]);
  msleep(DELAY);

  serialPutChar(packet_buffer[PACKET_ID_INDEX]);
  msleep(DELAY);

  serialPutChar(packet_buffer[OPCODE_INDEX]);
  msleep(DELAY);

  serialPutChar(packet_buffer[LENGTH_INDEX]);
  msleep(DELAY);

  for(index=0; index<packet_buffer[LENGTH_INDEX]; index++){
    serialPutChar(packet_buffer[DATA_INDEX + index]);
    msleep(DELAY);
  }

  serialPutChar(packet_buffer[CRC_INDEX]);
  msleep(DELAY);

  return true;
}

void sendResponse(){
  bool registering = false;
  packet_buffer[DESTINATION_ID_INDEX] = packet_buffer[SOURCE_ID_INDEX];

  if(packet_buffer[OPCODE_INDEX] == NO_OP){
    packet_buffer[OPCODE_INDEX] = ACKNOWLEDGEMENT;
    packet_buffer[LENGTH_INDEX] = 0;
  } else if(packet_buffer[OPCODE_INDEX] == REQUEST_NEW_ROBOT_IDS) {
    if(registered){
      packet_buffer[OPCODE_INDEX] = NO_OP;
      packet_buffer[LENGTH_INDEX] = 0;
    } else {
      packet_buffer[OPCODE_INDEX] = ACKNOWLEDGEMENT;
      packet_buffer[LENGTH_INDEX] = 0;
      registering = true;
    }
  } else if(packet_buffer[OPCODE_INDEX] == REQUEST_INFO) {
    char name[] = NAME;
    char index;
    packet_buffer[OPCODE_INDEX] = DATA;
    packet_buffer[LENGTH_INDEX] = INFO_SIZE;
    packet_buffer[DATA_INDEX] = SONAR_DATA_SIZE;
    for(index=0; index<INFO_SIZE-1; index++)
      packet_buffer[DATA_INDEX + index+1] = name[index];
  } else if(packet_buffer[OPCODE_INDEX] == REQUEST_SENSOR_LAYOUT) {
    char index;
    packet_buffer[OPCODE_INDEX] = DATA;
    packet_buffer[LENGTH_INDEX] = SENSOR_LAYOUT_SIZE;
    for(index=0; index<SENSOR_LAYOUT_SIZE; index++)
      packet_buffer[DATA_INDEX + index] = sensor_layout[index];
  } else if(packet_buffer[OPCODE_INDEX] == REQUEST_SENSOR_DATA) {
    char index;
    packet_buffer[OPCODE_INDEX] = DATA;
    packet_buffer[LENGTH_INDEX] = SENSOR_DATA_SIZE;
    for(index=0; index<SENSOR_DATA_SIZE; index++)
      packet_buffer[DATA_INDEX + index] = sensor_data[index];
  } else if(packet_buffer[OPCODE_INDEX] == SET_SENSE) {
    sense_enabled = (bool)packet_buffer[DATA_INDEX];
    packet_buffer[OPCODE_INDEX] = ACKNOWLEDGEMENT;
    packet_buffer[LENGTH_INDEX] = 0;
  } else {
    packet_buffer[OPCODE_INDEX] = NEGATIVE_ACKNOWLEDGEMENT;
    packet_buffer[LENGTH_INDEX] = 0;
  }

  if(packet_buffer[OPCODE_INDEX]!=NO_OP) {
    if(DEBUG) printf("sending...");
    sendPacket();
    if(DEBUG) printf("done\n");
    if(registering){
      if(receivePacket())
        if(packet_buffer[OPCODE_INDEX]==ACKNOWLEDGEMENT)
          registered = true;
    }
  } else
    if(DEBUG) printf("NO_OP");
}

void communicate() {
  while(1) {
    if(DEBUG) printf("receiving...");
    while(!receivePacket()){
      if(DEBUG) printf("*");
      defer();
    }
    if(DEBUG) printf("done\n");
    sendResponse();
  }
}

void sense(){
  int i, j;
  while(1){
    if(sense_enabled) {
      i = 0;
      sensor_data[i++] = analog(BREAK_BEAM_SENSOR_PORT);
      sensor_data[i++] = analog(LIGHT_SENSOR_PORT);
      sensor_data[i++] = digital(LEFT_FRONT_BUMP_SENSOR_PORT);
      sensor_data[i++] = digital(RIGHT_FRONT_BUMP_SENSOR_PORT);
    }
    defer();
  }
}

void sonar_scan(){
  bool clockwise = false;
  float curr_angle;
  float increment = 180.0/((float)(SONAR_DATA_SIZE - 1));
  long delay = (long)(DELAY_PER_DEGREE*increment);
  int i;

  while(1){
    if(sense_enabled) {
      if(clockwise){
        for(curr_angle=180.0, i=SONAR_DATA_END_INDEX;
            curr_angle>=0.0 && i>=SONAR_DATA_START_INDEX;
            curr_angle-=increment, i--){
          servo_deg(curr_angle);
          msleep(delay);
          sensor_data[i] = sonar_closeup();
          msleep(25l);
        }
        clockwise=false;
      } else {
        for(curr_angle=0.0, i=SONAR_DATA_START_INDEX;
            curr_angle<=180.0 && i<=SONAR_DATA_END_INDEX;
            curr_angle+=increment, i++){
          servo_deg(curr_angle);
          msleep(delay);
          sensor_data[i] = sonar_closeup();
          msleep(25l);
        }
        clockwise=true;
      }
    }
    defer();
  }
}

void main(void){
  int pid[3];

  init();

  printf("I am %s!\n", NAME);

  pid[0]=start_process(sense());
  pid[1]=start_process(communicate());
  pid[2]=start_process(sonar_scan());

  while(!stop_button()){
    defer();
  }

  kill_process(pid[0]);
  kill_process(pid[1]);
  kill_process(pid[2]);
  close();
}
